// Copyright (c) Facebook, Inc. and its affiliates.
//
// This source code is licensed under the MIT license found in the
// LICENSE file in the "hack" directory of this source tree.

use decl_provider::DeclProvider;
use hhbc_by_ref_ast_scope::{self as ast_scope, Scope};
use hhbc_by_ref_env::emitter::Emitter;
use hhbc_by_ref_hhbc_ast::SpecialClsRef;
use hhbc_by_ref_hhbc_ast::{ClassishKind, ClassishKind::*};
use hhbc_by_ref_hhbc_string_utils as string_utils;
use hhbc_by_ref_instruction_sequence::InstrSeq;
use naming_special_names_rust::classes;
use oxidized::{aast::*, ast, ast_defs};

#[derive(Debug)]
pub enum ClassExpr<'arena> {
    Special(SpecialClsRef),
    Id(ast_defs::Id),
    Expr(ast::Expr),
    Reified(InstrSeq<'arena>),
}

impl<'arena> ClassExpr<'arena> {
    fn get_original_class_name<'a, 'decl, D: DeclProvider<'decl>>(
        emitter: &Emitter<'arena, 'decl, D>,
        check_traits: bool,
        resolve_self: bool,
        opt_class_info: Option<(ClassishKind, &str)>,
    ) -> Option<String> {
        if let Some((kind, class_name)) = opt_class_info {
            if (kind != Trait || check_traits) && resolve_self {
                if string_utils::closures::unmangle_closure(class_name).is_none() {
                    return Some(class_name.to_string());
                } else if let Some(c) = emitter
                    .emit_global_state()
                    .get_closure_enclosing_class(class_name)
                {
                    if ClassishKind::from(c.kind.clone()) != Trait {
                        return Some(c.name.clone());
                    }
                }
            }
        }
        None
    }

    #[allow(clippy::needless_lifetimes)]
    pub fn get_parent_class_name<'a>(class: &ast_scope::Class<'a>) -> Option<String> {
        if let [Hint(_, hint)] = class.get_extends() {
            if let Hint_::Happly(ast_defs::Id(_, parent_cid), _) = &**hint {
                return Some(parent_cid.to_string());
            }
        }
        None
    }

    fn get_original_parent_class_name<'a, 'decl, D: DeclProvider<'decl>>(
        emitter: &Emitter<'arena, 'decl, D>,
        check_traits: bool,
        resolve_self: bool,
        opt_class_info: Option<(hhbc_by_ref_hhbc_ast::ClassishKind, &str)>,
        opt_parent_name: Option<String>,
    ) -> Option<String> {
        if let Some((kind, class_name)) = opt_class_info {
            if kind == hhbc_by_ref_hhbc_ast::ClassishKind::Interface {
                return Some(classes::PARENT.to_string());
            };
            if (kind != hhbc_by_ref_hhbc_ast::ClassishKind::Trait || check_traits) && resolve_self {
                if string_utils::closures::unmangle_closure(class_name).is_none() {
                    return opt_parent_name;
                } else if let Some(c) = emitter
                    .emit_global_state()
                    .get_closure_enclosing_class(class_name)
                {
                    return c.parent_class_name.clone();
                }
            }
        }
        None
    }

    fn expr_to_class_expr<'a, 'decl, D: DeclProvider<'decl>>(
        emitter: &Emitter<'arena, 'decl, D>,
        check_traits: bool,
        resolve_self: bool,
        scope: &Scope<'a, 'arena>,
        expr: ast::Expr,
    ) -> Self {
        if let Some(cd) = scope.get_class() {
            Self::expr_to_class_expr_(
                emitter,
                check_traits,
                resolve_self,
                Some((
                    hhbc_by_ref_hhbc_ast::ClassishKind::from(cd.get_kind()),
                    cd.get_name_str(),
                )),
                Self::get_parent_class_name(cd),
                expr,
            )
        } else {
            Self::expr_to_class_expr_(emitter, check_traits, resolve_self, None, None, expr)
        }
    }

    pub fn expr_to_class_expr_<'decl, D: DeclProvider<'decl>>(
        emitter: &Emitter<'arena, 'decl, D>,
        check_traits: bool,
        resolve_self: bool,
        opt_class_info: Option<(hhbc_by_ref_hhbc_ast::ClassishKind, &str)>,
        opt_parent_name: Option<String>,
        expr: ast::Expr,
    ) -> Self {
        match expr.2 {
            Expr_::Id(x) => {
                let ast_defs::Id(pos, id) = *x;
                if string_utils::is_static(&id) {
                    Self::Special(SpecialClsRef::Static)
                } else if string_utils::is_parent(&id) {
                    match Self::get_original_parent_class_name(
                        emitter,
                        check_traits,
                        resolve_self,
                        opt_class_info,
                        opt_parent_name,
                    ) {
                        Some(name) => Self::Id(ast_defs::Id(pos, name)),
                        None => Self::Special(SpecialClsRef::Parent),
                    }
                } else if string_utils::is_self(&id) {
                    match Self::get_original_class_name(
                        emitter,
                        check_traits,
                        resolve_self,
                        opt_class_info,
                    ) {
                        Some(name) => Self::Id(ast_defs::Id(pos, name)),
                        None => Self::Special(SpecialClsRef::Self_),
                    }
                } else {
                    Self::Id(ast_defs::Id(pos, id))
                }
            }
            _ => Self::Expr(expr),
        }
    }

    pub fn class_id_to_class_expr<'a, 'decl, D: DeclProvider<'decl>>(
        emitter: &Emitter<'arena, 'decl, D>,
        check_traits: bool,
        resolve_self: bool,
        scope: &Scope<'a, 'arena>,
        cid: &ast::ClassId,
    ) -> Self {
        let ClassId(_, annot, cid_) = cid;
        let expr = match cid_ {
            ClassId_::CIexpr(e) => e.clone(),
            ClassId_::CI(sid) => Expr((), annot.clone(), Expr_::mk_id(sid.clone())),
            ClassId_::CIparent => return Self::Special(SpecialClsRef::Parent),
            ClassId_::CIstatic => return Self::Special(SpecialClsRef::Static),
            ClassId_::CIself => return Self::Special(SpecialClsRef::Self_),
        };
        Self::expr_to_class_expr(emitter, check_traits, resolve_self, scope, expr)
    }
}
